sealed trait Tree[+A]
case class Leaf[A](value:A) extends Tree[A]
case class Branch[A](left:Tree[A], right:Tree[A]) extends Tree[A]

object Tree {    
    //Exercice 1
    def size[A](t:Tree[A]):Int = t match {
        case Leaf(_) => 1
        case Branch(l, r) => size(l) + size(r) + 1
    }

    //Exercice 2
    def maximum(t:Tree[Int]):Int = t match {
        case Leaf(v) => v
        case Branch(l, r) => maximum(l) max maximum(r)
    }

    //Exercice 3
    def depth[A](t:Tree[A]):Int = t match {
        case Leaf(_) => 0
        case Branch(l, r) => (depth(l) max depth(r)) + 1
    }

    //Exercice 4
    def map[A, B](t:Tree[A])(f:A => B):Tree[B] = t match {
        case Leaf(v) => Leaf(f(v))
        case Branch(l, r) => Branch(map(l)(f), map(r)(f))
    }

    //Exercice 5
    //Essai recursif terminal
    //def fold[A, B](t:Tree[A], z:B)(f(A, B) => B):B = t match {
    //    case Leaf(_) = f(_, z)
    //    case Branch(l, r) => fold(l, (fold(r, z)(f))(f))
    //}

    //Fonctionnel (et magnifique) mais non recursif terminal
    def fold[A, B](t:Tree[A])(f:A => B)(g:(B, B) => B):B = t match {
        case Leaf(v) => f(v)
        case Branch(l, r) => g(fold(l)(f)(g), fold(r)(f)(g))
    }

    //Réecritures avec le fold
    def sizeFold[A](t:Tree[A]):Int = fold(t)(a => 1)(1 + _ + _)
    def maximumFold(t:Tree[Int]):Int = fold(t)(a => a)(_ max _)
    def depthFold[A](t:Tree[A]):Int = fold(t)(a => 0)((l1, l2) => 1 + (l1 max l2))
    def mapFold[A, B](t:Tree[A])(f:A => B):Tree[B] = fold(t)(a => Leaf(f(a)):Tree[B])(Branch(_, _))
}